from .pager import *
from .section import *
import pandas as pd
import matplotlib

class experiment:

    def __init__(self, api, data):        
        if (not(data==None) & (type(data) == dict) & 
            ("experimentID" in data.keys()) &
            ("name" in data.keys())
           ):
            self.__api = api
            self.__experimentID = data["experimentID"]
            self.__name = data["name"]
            self.__title = "Experiment '"+self.__name+"' ("+str(self.__experimentID)+")"
            self.__data = data
        else:
            raise Exception("no (valid) experiment data") 
        
    def __repr__(self):
        return(self.__title)
        
    
    def sections(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self.__api,"Experiment Sections", 
                     "/api/v1/experiments/"+str(self.__experimentID)+"/sections", request, "expJournalID", "order", 5))
        
        
    def section(self, section_id):
        return(self.__api.section(section_id))
    
    def add(self, data, title=None, order=None):
        if data is None:
            return(None)
        elif type(data) == section:
            sectionType = data.type()
            if title==None:
                sectionHeader = data.title()
            else:
                sectionHeader = title
            sectionData = data.get()    
        else:
            if title==None:
                sectionHeader = "New section"
            else:
                sectionHeader = title
            if type(data) == pd.DataFrame:
                sectionType = "DATATABLE"
                sectionData = data
            elif type(data) == matplotlib.figure.Figure:    
                sectionType = "CANVAS"
                sectionData = data
            elif type(data) == PngImagePlugin.PngImageFile:   
                sectionType = "CANVAS"
                sectionData = data
            else:
                raise Exception("no (valid) section data")
        #create new section    
        request = {"sectionType": sectionType, "sectionHeader": sectionHeader}
        location = "/api/v1/experiments/"+str(self.__experimentID)+"/sections"
        section_id = self.__api.request(location, "post", json.dumps(request))       
        #get new section
        new_section = self.__api.section(section_id)
        #put content in section
        new_section.set(sectionData)
        #return new section
        return(new_section)
        
            
            
        
        


import pandas as pd
import math
import ipywidgets as widgets
from IPython.display import display, clear_output
from pandas.io.json import json_normalize


class pager:

    def __init__(self, api, title, location, request, index, sort, records):        
        #store
        self.__api = api
        self.__title = title
        self.__location = location
        self.__request = request
        self.__index = index
        self.__sort = sort
        self.__records = records
        self._reset()
        #get first page        
        self._set_page(self.__page)  
        
    def _reset(self):
        self.__page = 0
        self.__maxRecords = 0
        self.__totalRecords = 0
        self.__pages = 0
        self.__data = None
        
    def __repr__(self):
        description = self.__title
        if self.__pages>1:
            return(description+" ("+str(self.__pages)+" pages)")
        else:
            return(description)
        
    def _set_page(self, page):
        #define request
        request = self.__request
        request["$page"] = page
        request["$records"] = self.__records
        request["$sort"] = self.__sort
        rp = self.__api.request(self.__location,"get",request)
        #check and get
        if rp==None:
            self._reset()
        elif type(rp) == dict:
            if "maxRecords" in rp.keys():
                self.__maxRecords = rp["maxRecords"]
            else:
                raise Exception("unexpected response, no maxRecords")
            if "totalRecords" in rp.keys():
                self.__totalRecords = rp["totalRecords"]
            else:
                raise Exception("unexpected response, no totalRecords")    
            if "currentPage" in rp.keys():
                self.__page = rp["currentPage"]
            else:
                raise Exception("unexpected response, no currentPage")
            self.__pages = math.ceil(self.__totalRecords/self.__maxRecords)  
            if "data" in rp.keys():                        
                self.__page = page
                self.__data = pd.DataFrame(json_normalize(rp["data"]))
                if len(self.__data)>0:
                    self.__data = self.__data.set_index(self.__index)                
            else:
                raise Exception("unexpected response, no data")                    
        else:
            raise Exception("unexpected response, no dict")         
    
    
    #todo: reset button
    def show(self):   
        #data
        data = widgets.Output()                              
        #paging
        box_layout = widgets.Layout(
            display='flex', flex_flow='row', justify_content='space-between', 
            border='solid', width='100%')
        children = []
        header = widgets.Label()    
        button_previous = widgets.Button(description="previous", button_style="primary")                        
        button_next = widgets.Button(description="next", button_style="primary")                        
        if self.__pages>1:
            children = [button_previous, header, button_next]
        else:
            children = [widgets.Label(),header,widgets.Label()]
        box = widgets.Box(children=children, layout=box_layout)        
        #output
        display(widgets.VBox((box,data)))
            
        #set header content    
        def set_header_and_buttons(page):  
            if self.__pages>1:
                header.value=self.__title+" - page "+str(page+1)+" of "+str(self.__pages)
                if page==0:
                    button_previous.disabled=True
                    button_previous.button_style=""
                else:
                    button_previous.disabled=False
                    button_previous.button_style="primary"
                if page==(self.__pages-1):
                    button_next.disabled=True
                    button_next.button_style="" 
                else:
                    button_next.disabled=False
                    button_next.button_style="primary"    
            else:
                header.value=self.__title
                
        #define actions on buttons    
        def on_button_previous(b):
            if self.__page>0:
                set_header_and_buttons(self.__page-1) 
                button_previous.disabled=True
                button_next.disabled=True
                self._set_page(self.__page-1) 
                set_header_and_buttons(self.__page) 
                with data:
                    clear_output();
                    display(self.__data) 
                
        def on_button_next(b):
            if self.__page<(self.__pages-1):
                set_header_and_buttons(self.__page+1) 
                button_previous.disabled=True
                button_next.disabled=True
                self._set_page(self.__page+1) 
                set_header_and_buttons(self.__page) 
                with data:
                    clear_output();
                    display(self.__data) 
                
        set_header_and_buttons(self.__page)    
        with data:
            display(self.__data)  
        button_previous.on_click(on_button_previous)
        button_next.on_click(on_button_next)
        
        
        
        
        
        
        


from .pager import *
import json
import xlrd
import pandas as pd
import matplotlib
from PIL import Image, PngImagePlugin
from io import BytesIO


class section:

    def __init__(self, api, data):        
        if (not(data==None) & (type(data) == dict) & 
            ("expJournalID" in data.keys()) & 
            ("experimentID" in data.keys()) & 
            ("sectionType" in data.keys()) & 
            ("sectionHeader" in data.keys())
           ):
            self.__api = api
            self.__expJournalID = data["expJournalID"]
            self.__experimentID = data["experimentID"]
            self.__sectionType = data["sectionType"]
            self.__sectionHeader = data["sectionHeader"]
            self.__title = "Section '"+self.__sectionHeader+"' ("+str(self.__expJournalID)+", "+self.__sectionType+")"
            self.__data = data
        else:
            raise Exception("no (valid) section data") 
        
    def __repr__(self):
        return(self.__title)
        
    def type(self):
        return(self.__sectionType)
    
    def title(self):
        return(self.__sectionHeader)
    
    def experiment(self):
        request = {}
        return(self.__api.experiment(self.__experimentID))
           
    def get(self):
        if self.__sectionType=="DATATABLE":
            rp = self.__api.request("/api/v1/experiments/sections/"+str(self.__expJournalID)+"/datatable", "get", {})
            if not(rp==None) & (type(rp) == list):
                return(pd.DataFrame(rp))                                                 
            else:
                raise Exception("section type "+self.__sectionType+" returns unexpected data")
        elif self.__sectionType=="PARAGRAPH":
            rp = self.__api.request("/api/v1/experiments/sections/"+str(self.__expJournalID)+"/html", "get", {}, stream=True)
            return(rp.decode("utf-8"))                                                             
        elif self.__sectionType=="CANVAS":
            rp = self.__api.request("/api/v1/experiments/sections/"+str(self.__expJournalID)+"/canvas", "get", {}, stream=True)
            stream = BytesIO(rp)
            return(Image.open(stream))                                                             
        elif self.__sectionType=="EXCEL":
            rp = self.__api.request("/api/v1/experiments/sections/"+str(self.__expJournalID)+"/excel", "get", {}, stream=True)
            wb = xlrd.open_workbook(file_contents=rp)
            xlsx = pd.ExcelFile(wb, engine="xlrd")                        
            return(xlsx)                                                             
        elif self.__sectionType=="IMAGE":
            return(pager(self.__api, "Images", "/api/v1/experiments/sections/"+str(self.__expJournalID)+"/images", {}, "experimentFileID", "experimentFileID", 5))                                  
        else:
            raise Exception("section type "+self.__sectionType+" not supported")
            
    def set(self, data):
        if self.__sectionType=="DATATABLE":
            if type(data) == pd.DataFrame:
                location = "/api/v1/experiments/sections/"+str(self.__expJournalID)+"/datatable"
                rp = self.__api.request(location, "put", json.dumps(data.values.tolist()), headers={"Content-Type": "application/json"})                
            else:
                raise Exception("data type not supported to set "+self.__sectionType) 
        elif self.__sectionType=="CANVAS":
            location = "/api/v1/experiments/sections/"+str(self.__expJournalID)+"/canvas"
            if type(data) == matplotlib.figure.Figure:
                sizes = data.get_size_inches()
                dpi = 825/sizes[0]
                figfile = BytesIO()
                data.savefig(figfile, format="png", dpi=dpi, bbox_inches="tight")
                rp = self.__api.request(location, "put", figfile.getvalue(), headers={"Content-Type": "image/png"}) 
            elif type(data) == PngImagePlugin.PngImageFile:
                with BytesIO() as output:
                    data.save(output, format="png")
                    contents = output.getvalue()
                rp = self.__api.request(location, "put", contents, headers={"Content-Type": "image/png"}) 
            else:
                raise Exception("data type not supported to set "+self.__sectionType)    
        elif self.__sectionType=="EXCEL":
            if type(data) == pd.DataFrame:
                location = "/api/v1/experiments/sections/"+str(self.__expJournalID)+"/excel"
                raise Exception("section type EXCEL not (yet) supported")
                #rp = self.__api.request(location, "put", dump)                
            else:
                raise Exception("data type not supported to set "+self.__sectionType)                             
        else:
            raise Exception("section type "+self.__sectionType+" not supported")    
        
        
        


from .pager import *
from .experiment import *
from .section import *

import requests
import json
import keyring

def reset_key():
    keyring.delete_password("elabjournal-python", "apikey")   

class api:

    def __init__(self, key=None):
        
        self.__key = None
        self.__timeout = 180
        self.__url = "https://www.elabjournal.com"
        self.__user = None
        
        def check_and_set_key(key, throw_error, store_valid_key):
            if key==None:            
                return(False)
            elif (type(key) == str) & (len(key)>0):
                try:
                    rp = self.request("/api/v1/auth/user","get",{"body":{}},key)                   
                    if not(rp==None) & ("user" in rp.keys()):
                        print("Welcome "+rp["user"]["firstName"]+" "+rp["user"]["lastName"])
                        self.__key = key
                        self.__user = rp["user"]
                        if store_valid_key:                                                
                            keyring.set_password("elabjournal-python", "apikey", key)                
                        return(True) 
                    else:
                        return(False)
                except:
                    return(False)
            else:
                return(False)
        
        if key == None:
            keyring_key = keyring.get_password("elabjournal-python", "apikey")
            if not(check_and_set_key(keyring_key, False, False)):        
                from IPython.display import clear_output
                key = str(input("REST API key: "))
                clear_output();
                check_and_set_key(key, True, True)                                
        else:
          check_and_set_key(key, True, False)       
        
    
    def samples(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self, "Samples", "/api/v1/samples", request, "sampleID", None, 5))
    
    def sample_types(self, name=None, archived=None):
        request = {}
        if not(name==None):
            request["name"] = name
        if not(archived==None):
            request["archived"] = archived
        return(pager(self, "SampleTypes", "/api/v1/sampleTypes", request, "sampleTypeID", None, 5))
    
    def groups(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self, "Groups", "/api/v1/groups", request, "groupID", None, 5))
    
    def projects(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self, "Projects", "/api/v1/projects", request, "projectID", "projectID DESC", 5))
    
    def experiments(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self, "Experiments", "/api/v1/experiments", request, "experimentID", "experimentID DESC", 5))
    
    def studies(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self,"Studies", "/api/v1/studies", request, "studyID", "studyID DESC", 5))
    
    def storage_types(self, search=None):
        request = {}
        if not(search==None):
            request["search"] = search
        return(pager(self,"Storage Types", "/api/v1/storageTypes", request, "storageTypeID", "storageTypeID DESC", 5))
    
    def experiment(self, experiment_id):
        request = {"$expand": experiment_id}
        rp = self.request("/api/v1/experiments", "get", request)
        #check and get
        if not(rp==None) & (type(rp) == dict):
            if ("totalRecords" in rp.keys()) & (rp["totalRecords"]>=1):
                if "data" in rp.keys():
                    if(rp["totalRecords"]==1):
                        return(experiment(self, rp["data"][0]))
                    else:
                        for dataItem in rp["data"]:
                            if(dataItem["experimentID"]==experiment_id):
                                return(experiment(self, dataItem))
                        return(None)               
                else:   
                    return(None)
            else:
                return(None)                                                  
        else:
            return(None)
        
    def section(self, section_id):
        rp = self.request("/api/v1/experiments/sections/"+str(section_id), "get", {})
        #check and get
        if not(rp==None) & (type(rp) == dict):
            return(section(self,rp))                                                 
        else:
            return(None)    
    
    
    def __repr__(self):
        description = "eLABJournal API object"
        if self.__key == None:
            description += " - authentication failed"
        else:
            description += " - authenticated as "+self.__user["firstName"]+" "+self.__user["lastName"]
        return(description)    
        
    def request(self, location, method, request, key=None, show_messages=True, stream=False, headers=None):  
        if key==None:
            key = self.__key
        try: 
            request_headers = {"Accept": "application/json"}
            request_headers.update({"Authorization": key})    
            if not(headers==None):
                request_headers.update(headers)
            if method=="get":
                if type(request) == dict:
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.get(self.__url+location, params=data, timeout=self.__timeout, headers=request_headers, stream=stream)
            elif method=="post":    
                if type(request) == str:
                    request_headers.update({"Content-Type": "application/json"})
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.post(self.__url+location, data=data, timeout=self.__timeout, headers=request_headers, stream=stream)                
            elif method=="put":    
                if (type(request) == str) | (type(request) == bytes):
                    data = request
                else:
                    raise Exception("unsupported type of request") 
                response = requests.put(self.__url+location, data=data, timeout=self.__timeout, headers=request_headers, stream=stream)                
            else:
                raise Exception("unsupported method")   
            response.raise_for_status() 
            #no content
            if response.status_code == 204:
                return(None)
            elif stream:
                return(response.content)
            else:    
                try:
                    return(json.loads(response.text))
                except json.JSONDecodeError as decodeErr:
                    print ("JSON Decode Error:",decodeErr) 
                    raise Exception("no (valid) response from eLABJournal")            
        except requests.exceptions.HTTPError as httpErr: 
            if show_messages:
                print ("Http Error:",httpErr) 
            try:
                return(json.loads(response.text))
            except json.JSONDecodeError as decodeErr:
                return(None)           
        except requests.exceptions.ConnectionError as connErr: 
            if show_messages:
                print ("Error Connecting:",connErr) 
            return(None)
        except requests.exceptions.Timeout as timeOutErr: 
            if show_messages:
                print ("Timeout Error:",timeOutErr) 
            return(None)
        except requests.exceptions.RequestException as reqErr: 
            if show_messages:
                print ("Something Else:",reqErr) 
            return(None)
        raise Exception("no (valid) response from eLABJournal")     
        
        


from setuptools import setup, find_packages

setup(
    name="elabjournal",
    version="0.0.6",
    author="Matthijs Brouwer",
    url="https://github.com/matthijsbrouwer/elabjournal-python",
    packages=find_packages(),
    license="Apache License 2.0",
    long_description=open("README.txt").read(),
    install_requires=[
        "requests >= 2.18.4",
        "pandas",
        "keyring",
        "IPython",
        "rjson",
        "xlrd",
        "matplotlib",
        "Pillow",
        "ipywidgets"
    ],
    classifiers=[
        "Framework :: Jupyter",
        "Intended Audience :: Science/Research",
    ],
)

